%% ========================================================================
% Test of LPempc: Energy system case study (closed-loop)
% Written by: Leo Emil Sokoler, leso@imm.dtu.dk
% 02/05/2013, DTU Informatics
% =========================================================================
clear; clc; close all;
rng('default');

%% ========================================================================
% SYSTEM PARAMTERS
% =========================================================================
m = 3;                      % nr. of units
Tau = 30 + rand(m,1)*50/2;  % time constants

%% ========================================================================
% SYSTEM SETUP
% =========================================================================
Ts = 5;
[A,B,E,Cy,Cz] = ESsystem(Tau,Ts);
[nx,nu] = size(B);
ny = size(Cy,1);
nd = size(E,2);
nz = size(Cz,1);
x0_plant = zeros(nx,1);
u_m1 = zeros(nu,1);

%% ========================================================================
% CONTROLLER SETUP
%  ========================================================================
% paramters
N = 40;                         % prediction horizon
p = 1./Tau;                     % input costs
q = 1e3*ones(nz,1);             % soft constraints penalty
u_min = zeros(nu,1);            % min input
u_max = 200*rand(m,1) + 20;     % max input
Du_min = -20*rand(m,1) - 20;    % min input rate
Du_max = 20*rand(m,1) + 20;     % max input rate

%% ========================================================================
% SCNEARIO SETUP
% =========================================================================
M = 400;
T = 0:Ts:Ts*M;                 % time vector
P = zeros(M,1);
% ramp
idx = floor(N/4);
P(1:idx) = linspace(0,20,idx);
P(idx:end) = 20*ones(length(P(idx:end)),1);
% step
idx = floor(M/2);
P(idx:end) = 40*ones(length(P(idx:end)),1);
P = [P; repmat(P(end),N,1)];

%% ========================================================================
%  KALMAN FILTER SETUP
%  ========================================================================
x_plant = zeros(nx,M+1);
y_plant = zeros(ny,M+1);
z_plant = zeros(nz,M+1);
u = zeros(nu,M+1);

% NOISE SETUP
Rd = eye(nd);
Re = eye(ny);
sigma = 0.1;
dbar = zeros(nd,1);
ebar = zeros(ny,1);
d = chol(Rd)*randn(nd,M)*sigma;
e = chol(Re)*randn(ny,M)*sigma;
v = E*d;
Rv = E*Rd*E';

% PREALLOCATION FOR KALMAN FILTER
x_pred = zeros(nx,M+1);
y_pred = zeros(ny,M+1);
y_perr = zeros(ny,M+1);
P_pred = zeros(nx,nx,M+1);
x_inf = zeros(nx,M+1);
y_inf = zeros(ny,M+1);
y_ierr = zeros(ny,M+1);
P_inf = zeros(nx,nx,M+1);
kappa = zeros(nx,ny,M+1);

% INITIALIZE VALUES FOR KALMAN FILTER
x_inf(:,1) = x0_plant;
y_inf(:,1) = Cy*x_inf(:,1);
y_ierr(:,1) = y_plant(:,1) - y_inf(:,1);
P_inf(:,:,1) = zeros(nx);

% INITIALIZE PLANT
x_plant(:,1) = x0_plant;
y_plant(:,1) = Cy*x_plant(:,1) + e(:,1);
z_plant(:,1) = Cz*x_plant(:,1);

% PROBLEM DIMENSIONS
m_I = N*(4*nu+3*nz);
m_E = N*nx;
n = N*(nu+nz+nx);

% MIN/MAX LIMITS
Z_min = max(P - 5,0);
Z_max = P + 5;

% SIMULATE
for i = 1:M
    % MPC CONTROLLER
    z_min = Z_min(i:i+N-1)';
    z_max = Z_max(i:i+N-1)';

    % SOLVE PROBLEM
    [u_mpc,x_mpc,y_mpc,rho_mpc,info,output] = LPempc(A,B,E,Cz,dbar,x_inf(:,i),u_m1,u_min, ...
        u_max,Du_min,Du_max,z_min,z_max,p,q,N);
    
    % EXTRACT SOLUTION
    u(:,i) = u_mpc(:,1);
    u_m1 = u_mpc(:,1);
    
    % KALMAN FILTER PREDICTION
    x_pred(:,i) = A*x_inf(:,i) + B*u(:,i);
    y_pred(:,i) = Cy*x_pred(:,i);
    P_pred(:,:,i) = A*P_inf(:,:,i)*A' + Rv;
    
    % PLANT UPDATE
    x_plant(:,i+1) = A*x_plant(:,i) + B*u(:,i) + v(:,i);
    y_plant(:,i+1) = Cy*x_plant(:,i+1) + e(:,i);
    z_plant(:,i+1) = Cz*x_plant(:,i+1);
    
    % KALMAN FILTER INFERENCE
    kappa(:,:,i+1) = P_pred(:,:,i)*Cy'/(Cy*P_pred(:,:,i)*Cy' + Re);
    y_perr(:,i+1) = y_plant(:,i+1) - y_pred(:,i);
    x_inf(:,i+1) = x_pred(:,i) + kappa(:,:,i+1)*y_perr(:,i+1);
    y_inf(:,i+1) = Cy*x_inf(:,i+1);
    y_ierr(:,i+1) = y_plant(:,i+1) - y_inf(:,i+1);
    P_inf(:,:,i+1) = P_pred(:,:,i) - ...
        P_pred(:,:,i)*Cy'/(Cy*P_pred(:,:,i)*Cy' + Re)*Cy*P_pred(:,:,i);
end
i = i + 1;
% KALMAN FILTER PREDICTION
x_pred(:,i) = A*x_inf(:,i) + B*u(:,i);
y_pred(:,i) = Cy*x_pred(:,i);
P_pred(:,:,i) = A*P_inf(:,:,i)*A' + Rv;

%% ========================================================================
% PLOT TOTAL POWER PRODUCTION
% =========================================================================
subplot(2,1,1);
plot(T,Cz*x_plant); grid on;  hold on;
plot(T(2:end),Z_min(1:M), 'r--'); 
plot(T(2:end),Z_max(1:M), 'r--');
ylabel('C_z');
subplot(2,1,2);
stairs(T,u'); grid on;
ylabel('U');
xlabel('time [sec]'); 

