function [u,x,y,rho,info,output] = LPempc(A,B,E,C,d,x0,u_m1,u_min, ...
    u_max,Du_min,Du_max,y_min,y_max,p,q,N,kmax,tol,init,nu)
% LPempc: Riccati-based Interioir-Point Algorithm for Linear Economic
% Model Predictive Control.
% -------------------------------------------------------------------------
% INFO
% -------------------------------------------------------------------------
% Solves the Optimal Control Problem:
% min.   sum(p_k u_k + q_{k+1} rho_{k+1},k = 0,1,...,N)
% s. t.  x_{k+1} = A x_k + B u_k + E d_k,                k = 0,1,...,N-1
%        y_k = C x_k,                                    k = 1,2,...,N
%        (u_min)_k  <= u_k <= (u_max)_k,                 k = 0,1,...,N-1
%        (Du_min)_k <= u_k - u_{k-1} <= (Du_max)_k,      k = 0,1,...,N-1
%        (y_min)_k - rho_k <= y_k <= (y_max)_k + rho_k,  k = 1,2,...,N
%        rho_k >= 0,                                     k = 1,2,...,N
%
% Stan_dard Primal/Dual Form:
% (P) min_{t}   {g^T t | Ft = b, Ht + s = c, s >= 0}
% (D) max_{v,w} {-b^T v - c^T w | -F^T v - H^T w = g, w >= 0}
%
% -------------------------------------------------------------------------
% SYNTAX
% -------------------------------------------------------------------------
% [u,x,y,rho,info,output] = LPempc(A,B,E,C,d,x0,u_m1,u_min,u_max, ...
%       Du_min,Du_max,y_min,y_max,p,q,N)
%
% [u,x,y,rho,info,output] = LPempc(A,B,E,C,d,x0,u_m1,u_min,u_max, ...
%       Du_min,Du_max,y_min,y_max,p,q,N,kmax,tol,init,n_u)
%
% -------------------------------------------------------------------------
% INPUTS
% -------------------------------------------------------------------------
% (A,B,E,C)      : state space matrices   [(nx,nx),(nx,nu),(nx,nd),(ny,nx)]
% d              : disturbance            [(nd,N) or (nd,1)]
% x0             : initial state          [(nx,1)]
% u_m1           : previous input         [(nu,1)]
% u_min/u_max    : input limits           [(nu,N) or (nu,1)]
% Du_min/Du_max  : input-rate limits      [(nu,N) or (nu,1)]
% y_min/y_max    : output limits          [(ny,N) or (ny,1)]
% p              : input prices           [(nu,N) or (nu,1)] 
% q              : output violation price [(ny,N) or (ny,1)]
% N              : prediction horizon       
% n_u             : boun_dary parameter
% kmax (optional): max nr. of iterations
% tol  (optional): tolerance struct consisting of tol.E, tol.I, tol.D,
%                    tol.G, tol.Tau an_d tol.O
% init (optional): initial guess structure (init.t0, init.v0, init.s0,
%                    init.w0, init.tau0, init.kappa0)
%
% -------------------------------------------------------------------------
% OUTPUTS
% -------------------------------------------------------------------------
% u              : optimal input sequence   [u0, u1, ..., u_{N-1}]
% x              : optimal state sequence   [x1, x2, ..., x_N]
% y              : optimal output sequence  [y1, y2, ..., y_N]
% rho            : optimal penalty sequence [rho1, rho2, ..., rho_N]
% info           : info struct consisting of
%                    info.iter ~ n_umber of iterations
%                    info.status == 1 => converged,
%                    info.status == 0 => max. nr of iterations exceeded
%                    info.status == -1 => primal infeasible
%                    info.status == -2 => dual infeasible
%                    info.cputime ~ cputime
% output         : optimal values (t^*, v^*, w^*, s^*,tau^*, kappa^*)

% -------------------------------------------------------------------------
% REFERENCE & FURTHER INFO
% -------------------------------------------------------------------------
% Real-Time Optimization for Economic Model Predictive Control.
% Sokoler, Leo Emil; Edlun_d, Kristian; Frison, Gianluca; Skajaa, An_ders;
% Jrgensen, John Bagterp. In: The 10th European Workshop on Advanced
% Control an_d Diagnosis (ACD 2012). Technical University of Denmark, 2012.
%
% ========================================================================
% Leo Emil Sokoler (leso@dtu.dk), 30.04.2013
% DTU Informatics
% ========================================================================

%% ========================================================================
%  PROBLEM DIMENSIONS
%  ========================================================================
[n_x,n_u] = size(B);
n_d = size(E,2);
n_y = size(C,1);
n_xh = n_x + n_u;
m_I = N*(4*n_u+3*n_y);
m_E = N*n_xh;
n = N*(n_u+n_y+n_xh);

%% ========================================================================
%  DEFAULT INPUTS
%  ========================================================================
% max nr. of iterations
if(nargin < 17 || isempty(kmax))
    kmax = 100;
end

% tolerance settings
if(nargin < 18 || isempty(tol))
    tolE = 1e-8;
    tolI = 1e-8;
    tolD = 1e-8;
    tolG = 1e-8;
    tolO = 1e-8;
    tolTau = 1e-10;
else
    try
        tolE = tol.E;
        tolI = tol.I;
        tolD = tol.D;
        tolG = tol.G;
        tolTau = tol.Tau;
        tolO = tol.O;
    catch err
        warning(err.identifier);
        disp(['error in tolerance specifications, using default' ...
            ' tolerance settings']);
        tolE = 1e-8;
        tolI = 1e-8;
        tolD = 1e-8;
        tolG = 1e-8;
        tolO = 1e-8;
        tolTau = 1e-10;
    end
end

% initial guess
if(nargin < 19 || isempty(init))
    ti = zeros(n,1);
    vi = zeros(m_E,1);
    wi = ones(m_I,1);
    si = ones(m_I,1);
    kappai = 1;
    taui = 1;
else
    try
        ti = init.t;
        vi = init.v;
        wi = init.w;
        si = init.s;
        kappai = init.kappa;
        taui = init.tau;
    catch err
        warning(err.identifier);
        disp(['error in initial point definition, ' ...
            'using deafult initial point']);
        ti = zeros(n,1);
        vi = zeros(m_E,1);
        wi = ones(m_I,1);
        si = ones(m_I,1);
        kappai = 1;
        taui = 1;
    end
    
end

% max nr. of iterations
if(nargin < 20 || isempty(n_u))
    nu = 0.9995;
end

% append to problem data if necessary
if(size(y_min,2) < N)
    y_min = kron(ones(1,N),  y_min(:));
end

if(size(y_max,2) < N)
    y_max = kron(ones(1,N),  y_max(:));
end

if(size(u_min,2) < N)
    u_min = kron(ones(1,N),  u_min(:));
end

if(size(u_max,2) < N)
    u_max = kron(ones(1,N),  u_max(:));
end


if(size(Du_min,2) < N)
    Du_min = kron(ones(1,N), Du_min(:));
end

if(size(Du_max,2) < N)
    Du_max = kron(ones(1,N),  Du_max(:));
end

if(size(p,2) < N)
    p = kron(ones(1,N),  p(:));
end

if(size(q,2) < N)
    q = kron(ones(1,N),  q(:));
end

if(size(d,2) < N)
    d = kron(ones(1,N), zeros(n_d,1));
end


%% ========================================================================
%  INPUT CHECKS
%  ========================================================================
if(length(kmax(:)) ~= 1 || ~isreal(kmax(:))) 
    error('inconsistent dimensions: kmax')
end

if(length(n_u(:)) ~= 1 || ~isreal(n_u(:))) 
    error('inconsistent dimensions: n_u')
end

if(length(tolE(:)) ~= 1 || ~isreal(tolE(:))) 
    error('inconsistent dimensions: tolE')
end

if(length(tolI(:)) ~= 1 || ~isreal(tolI(:))) 
    error('inconsistent dimensions: tolI')
end

if(length(tolD(:)) ~= 1 || ~isreal(tolD(:))) 
    error('inconsistent dimensions: tolD')
end

if(length(tolG(:)) ~= 1 || ~isreal(tolG(:))) 
    error('inconsistent dimensions: tolG')
end

if(length(tolO(:)) ~= 1 || ~isreal(tolO(:))) 
    error('inconsistent dimensions: tolO')
end

if(length(tolTau(:)) ~= 1 || ~isreal(tolTau(:))) 
    error('inconsistent dimensions: tolTau')
end

if(length(kmax(:)) ~= 1) 
    error('inconsistent dimensions: kmax')
end


if(size(A,1) ~= n_x || size(A,2) ~= n_x || size(B,1) ~= n_x || ...
        size(B,2) ~= n_u || size(C,1) ~= n_y || size(C,2) ~= n_x ...
        || size(E,1) ~= n_x || size(E,2) ~= n_d)
    error('inconsistent dimensions: state space matrices')
end

if(size(p,1) ~= n_u || size(p,2) ~= N)
    error('inconsistent dimensions: input prices')
end

if(size(q,1) ~= n_y || size(q,2) ~= N)
    error('inconsistent dimensions: output violation prices')
end

if(size(y_min,1) ~= n_y || size(y_max,1) ~= n_y  || ...
        size(y_min,2) ~= N || size(y_max,2) ~= N)
    error('inconsistent dimensions: output limits')
end

if(size(u_min,1) ~= n_u || size(u_max,1) ~= n_u || ...
        size(u_min,2) ~= N || size(u_max,2) ~= N)
    error('inconsistent dimensions: input limits')
end

if(size(Du_min,1) ~= n_u || size(Du_max,1) ~= n_u || ...
        size(Du_min,2) ~= N || size(Du_max,2) ~= N)
    error('inconsistent dimensions: input limits')
end

if(size(d,1) ~= n_d || size(d,2) ~= N)
    error('inconsistent dimensions: disturbance')
end

if(size(u_m1,1) ~= n_u || size(u_m1,2) ~= 1)
    error('inconsistent dimensions: previous input')
end

if(size(x0,1) ~= n_x || size(x0,2) ~= 1)
    error('inconsistent dimensions: initial state')
end

if(size(ti,1) ~= N*(n_u+n_y+n_xh) || size(ti,2) ~= 1 || ...
        size(vi,1) ~= N*n_xh         || size(vi,2) ~= 1 || ...
        size(wi,1) ~= N*(4*n_u+3*n_y) || size(wi,2) ~= 1 || ...
        size(si,1) ~= N*(4*n_u+3*n_y) || size(si,2) ~= 1 || ...
        ~all(size(taui) == [1 1]) || ~all(size(taui) == [1 1]))
    error('inconsistent dimensions: initial solution')
end

%% ========================================================================
%  CALL MEX-FILE
%  ========================================================================
[t,v,w,s,tau,kappa,iter,status,cputime] = mxLPempc(A, B, E, C, d, x0, ...
    u_m1, u_min, u_max, Du_min, Du_max, y_min, y_max, p, q, N, ti, vi, ...
    wi, si, taui, kappai, kmax, tolE, tolI, tolD, tolG, tolO, tolTau, nu);

%% ========================================================================
%  STRUCTURE OUTPUT
%  ========================================================================
X = reshape(t/tau,n_u+n_xh+n_y,N);
u = X(1:n_u,:);
x = X(n_u+1:n_u+n_x,:);
y = C*x;
rho = X(n_u+n_xh+1:n_u+n_xh+n_y,:);

%% ========================================================================
%  WRITE TO OUTPUT
%  ========================================================================
info.iter = iter;
info.status = status;
info.cputime = cputime;
output.t = t;
output.v = v;
output.w = w;
output.s = s;
output.tau = tau;
output.kappa = kappa;

